#! /usr/bin/perl 

# Copyright (c) 2005,2006 ARM <openbsd -at- otterhole.ca>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# If you use this software, please send a note to the above email
# address indicating anything interesting you have to share.  This
# is a request and imposses no further obligation.


sub how () {
print " progname [-v] [-s] [-t <n>] [-f <n>] [-l <n>] [age]
      -v   -> verbose, twice changes verbosity
      -h   -> print this page
      -A   -> align
      -S   -> signatures (much quicker w/o)
      -t # -> show top # hits, includes
              # is number of records in the list
      -T # -> show top # hits, includes
              # is number of ports in a port list
      -f # -> use filter number #.  Filters are
              coded in the script.
      -F # -> supply your own filter...
      -r # -> read input file
      -l # -> include only entries with hits in the
              last # seconds
      -d # -> # of secs diff between first and last
      -s # -> # 0=deny,  1=pass, 2=pass or deny (default)
      -i # -> # interface
      -I # -> # not interface
      -p # -> # period (month, day, )
      -m # -> # mulitplier scale
      -a # -> # force architecture 'o35', 'o36', 'csv1'
              Not all features work for all architectures...
";
 exit;
}

########################################################################
# Known Bugs/TODO, etc:
#  - byte counts don't always work, even when info is available
#  - Filter and filter only currently work with pcap/tcpdump input
#  - other filters could and should work with tcpdump filter, but
#    are implemented as text filters
#  - more settings, perhaps args, should be identified
#
# Tue Jun 29 2004: ARM
#   Took old pf reporting tool and made something more useful
#   - reduced output to only the multiple port table by default
#     (old behaviour by using -v)
#   - added checking of signatures
#     (disabled using -s)
#   - added "text graph" to display distribution of hits over
#     the day
#   - cleaned up a bit, although still pretty ugly
#   - added a dynamic limit to the number entries per table (-t)
# arguments must be given in the correct order, oops.
#     progname [-v] [-s] [-t <n>] [age]
#
# Thu Jul 15 18:58:45 UTC 2004
#   - added [-l <last>] option
#
# Fri Jul 30 18:15:00 UTC 2004
#   - added -f <n> argument
#   - corrected time alignment
#
# Thu Jan 20 22:56:56 EST 2005
#   - added support for user chosen filter on command line
#   - added support for reading arbitrary files
#   - added support to calculate byte count
#
# Tue Apr 26 14:20:28 EDT 2005
#   - lots of things done over the past months, not identified
#
# Wed Aug 10 22:42:33 EDT 2005
#   - tried to clean up and better document variables
#   - only populate some variables if output will be used, reducing memory
# Sat Aug 13 12:37:16 EDT 2005
#   - cleaned up some of the use of variables, still ugly...
#   - added monthly reports
#   - added a scale option
#   - added ability to process multiple types of input
#     not all features work for all input format types...
# Thu Sep  1 18:32:54 EDT 2005
#   - added a filter for http style 'access' input log type
#   - added a flag for the 'align' value
#
# Fri Mar 17 08:33:08 EST 2006
#  learned of /usr/src/share/misc/license.template
#  and made it apply
#
########################################################################

use strict;
use Time::Local;
use POSIX;

my $ntop=30;    # -T Report the top N items
my $nthresh=10; # -t Report the top N items
my $archive=0;
my $cmd;        # Set later as the command called to get log data
my $align=1;    # something for TG
my $lastdate;
my $firstdate;  # First found time in the reocrds, in epoch time
my $ifirstdate; # First found time in the reocrds, in epoch time
my $lastaddr="";# Last address, for records on more than one line.
my $now=timelocal(localtime);
my $n;
my %tgq;      # TG for $quad
my %tgm;      # main TG
my $inlastsecs=0;
my $flagquiet=1;
my $flagsigs="";
my $infile="";
my $infiletype="pcap";
#y $duration='month';
my $duration='day';
my $diff;       # report only on IPs with records over this perid of time
my $state=2;    # pass, deny, etc.
my $intf;       # look only for records on this interface
my $nintf;      # look only for records not on this interface
my $scale=1;      # Scale output
my %regex;
my %regexstart;
my $args;

my %bysa;       # By Source Addr
                #  {'count'} = packet count
                #  {'first'} = first timestamp
                #  {'last'} = last timestamp
                #  {'Nets'} = ref. to hash of dstaddr
                #  {'Ips'} = ref. to hash of dstport
                #  {'bytes'} = byte count
my %byport;     # By port
                #  {'count'} = packet count
                #  {'bytes'} = byte count
                #  {'Nets'} = ref. to hash of dstnet
                #  {'Ips'} = ref. to hash of dstaddr
my %byquad;     # By srcaddr-dstaddr-dstport 
                #  {'count'} = packet count
                #  {'last'} = last timestamp
my %total;      # Totals for the entire review
                #  {'count'} = packet count
                #  {'bytes'} = byte count
my %bynet;      # By src network
                #  {'count'} = packet count
                #  {'bytes'} = byte count

## added for trace files with dates...
my %M = ( "Jan", 0,
          "Feb", 1,
          "Mar", 2,
          "Apr", 3,
          "May", 4,
          "Jun", 5,
          "Jul", 6,
          "Aug", 7,
          "Sep", 8,
          "Oct", 9,
          "Nov",10,
          "Dec",11,
         );

my   %sigs=( "s41" , "4141 4141 4141 4141 4141 4141 4141 4141",
             "s45" , "4545 4545 4545 4545 4545 4545 4545 4545",
             "s4x" , "4546 4748 494a 4b4c 4d4e 4f50 5152 5354",
             "swp" , "6566 6768 696a 6b6c 6d6e 6f70 7172 7374",
             "shp" , "[^ 	] ac11 00.. ",
            );

my $fw="o36";

########################################################################
# Setup Pre-Defined filters
########################################################################
my $nfilter=0;
my $Filter;
# it might be nicer to use a hash and have filter rules with names
# Ips should be set to appropriate values for your network
my $DarkNet="10.254.198.0/24";
my $MyIP="10.243.1.1";
my $MyCorp =  "    src net 10.1.0.0/16  or src net 10.10.0/12".
              " or src net 10.135.64.0/18 or src net 10.198.253.0/24" ;
my $MyAS   =  "    src net 47.0.0.0/14" .
              " or src net 47.47.0.0/15" ;
my @filter;
#  $filter[1]="dst net " . $DarkNet ;
#  $filter[2]="net " . $DarkNet;
#  $filter[3]="host " . $MyIP;
#  $filter[4]="dst host " . $MyIP;
#  $filter[5]="src host " . $MyIP;
#  $filter[7]=$MyCorp;
#  $filter[8]=$MyCorp." or ".$MyAS;
#  $filter[9]="dst net ".$DarkNet." and \\(".$MyCorp." or ".$MyAS."\\)";

   $filter[0]="not src net 172.16.0.0/24 and not dst host 172.16.0.10"
             . " and not dst host 172.17.0.11 and not dst host 172.16.0.12" ;
   $filter[1]="not src net 172.16.0.0/24" ;
   $filter[2]="not src net 172.16.0.0/24 and port 80" ;
   $filter[3]="not src net 172.16.0.0/24 and dst net $DarkNet" ;
   $filter[4]="src net 172.16.22.0/24" ;






########################################################################
# Process arguments
########################################################################

my $args=join(" ", @ARGV);
if ($#ARGV > -1 && $ARGV[0] eq "-v" ) { $flagquiet--;shift } 
if ($#ARGV > -1 && $ARGV[0] eq "-v" ) { $flagquiet--;shift } 
if ($#ARGV > -1 && $ARGV[0] eq "-v" ) { $flagquiet--;shift } 
if ($#ARGV > -1 && $ARGV[0] eq "-A" ) { $align=0;shift;} 
if ($#ARGV > -1 && $ARGV[0] eq "-h" ) { &how } 
if ($#ARGV > -1 && $ARGV[0] eq "-S" ) { $flagsigs="-x";shift } 
if ($#ARGV > -1 && $ARGV[0] eq "-t" ) { $ntop=$ARGV[1];shift;shift} 
if ($#ARGV > -1 && $ARGV[0] eq "-T" ) { $nthresh=$ARGV[1];shift;shift} 
if ($#ARGV > -1 && $ARGV[0] eq "-f" ) { $nfilter=$ARGV[1];shift;shift} 
if ($#ARGV > -1 && $ARGV[0] eq "-F" ) { $nfilter=-1;$Filter=$ARGV[1];shift;shift} 
if ($#ARGV > -1 && $ARGV[0] eq "-r" ) { $infile=$ARGV[1] ;shift;shift} 
if ($#ARGV > -1 && $ARGV[0] eq "-l" ) { $inlastsecs=$ARGV[1]  ;shift;shift} 
if ($#ARGV > -1 && $ARGV[0] eq "-d" ) { $diff=$ARGV[1]  ;shift;shift} 
if ($#ARGV > -1 && $ARGV[0] eq "-s" ) { $state=$ARGV[1]  ;shift;shift} 
if ($#ARGV > -1 && $ARGV[0] eq "-i" ) { $intf=$ARGV[1]  ;shift;shift} 
if ($#ARGV > -1 && $ARGV[0] eq "-I" ) { $nintf=$ARGV[1]  ;shift;shift} 
if ($#ARGV > -1 && $ARGV[0] eq "-p" ) { $duration=$ARGV[1]  ;shift;shift} 
if ($#ARGV > -1 && $ARGV[0] eq "-m" ) { $scale=$ARGV[1]  ;shift;shift} 
if ($#ARGV > -1 && $ARGV[0] eq "-a" ) { $fw=$ARGV[1]  ;shift;shift} 
if ( $fw eq "csv1" ) { $infiletype="csv"; }
if ( $fw eq "f48" )  { $infiletype="txt"; }
if ( $fw eq "acc" )  { $infiletype="txt"; }

my $filter;
if ($nfilter != -1 ) {
  $filter=$filter[$nfilter];
} else {
  $filter=$Filter;
} 

my $tcpdump="/usr/sbin/tcpdump -nttvr";
 if ($state != 2 || $intf || $nintf) {
    $tcpdump="/usr/sbin/tcpdump -nettvr";
 }
if ($#ARGV > -1 ) { 
 $archive=1;
 $cmd="/usr/bin/gzcat /var/log/pflog." . $ARGV[0] . 
       ".gz|$tcpdump - $flagsigs $filter|";
} else {
 $cmd="$tcpdump /var/log/pflog $flagsigs $filter|";
}
if ($infile) {
 $archive=1;
 if ($infile =~ /z$/i ) {
  print "INFILE: $infile\n";
  if ( $infiletype eq "csv" || $infiletype eq "txt" ) {
    $cmd="/usr/bin/zcat $infile|";
  } else {
    $cmd="/usr/bin/gzcat $infile| $tcpdump - $flagsigs $filter|";
  }
 } else {
  if ( $infiletype eq "csv" || $infiletype eq "txt") {
    $cmd="/bin/cat $infile |";
  } else {
    $cmd="$tcpdump $infile $flagsigs $filter|";
  }
 } 
}

########################################################################
# TG Text graph global settings.  These should not be global....
########################################################################
my %tg;
my $tgunit;
my $tgunits;
my $tgtextd;
   my $tgduration= 60 * 60 * 24  ;
#( $tgunit, $tgunits,$tgtextd ) = &tgautoset($tgduration); ## TG
if ($duration eq 'month' ) {
 ( $tgunit, $tgunits,$tgtextd ) = ( 60 * 60 *24 , 32, "day");
} else {
 ( $tgunit, $tgunits,$tgtextd ) = ( 60 * 60 , 24, "hour");
}

########################################################################
# Start Processing
########################################################################
open(LOG , $cmd) or die;


$regexstart{'o35'}='^(\d+.\d+)';
$regexstart{'o35s'}='^(\d+.\d+)';
$regexstart{'o36'}='^(\d+.\d+)';
$regexstart{'o36s'}='^(\d+.\d+)';
$regexstart{'csv1'}='^"([^"]*?)" "([^"]*?)" "([^"]*?)" ';
$regexstart{'f48'}='^(\S+)\s+(\d+)\s+\d+:\d+:\d+.*?:\s(\d+):(\d+):(\d+)';
$regexstart{'acc'}='^\S+\s+\S+\s+\S+\s+\[(\d+)\/(\S+)\/(\d+):(\d+):(\d+):(\d+)\s';
#204.101.134.87 - - [09/Nov/2004:14:06:57 -0500] "GET / HTTP/1.0" 200 641
while (<LOG>) {
 if (/^$regexstart{$fw}/) { 
#print "HELLO $_";
  # next if ($1 =~ /^Number$/);
    next if ($1 eq "Number");
    my $mynow = $1;
    if ( $fw eq "csv1" ) {
      my $mydate=$2;
      my $walltime=$3;
      my @mydate = $mydate =~ /^(\d+)(\D+)(\d+)$/;
      my @walltime= $walltime =~ /^(\d+):(\d+):(\d+)$/;
         $mynow=mktime($walltime[2],$walltime[1],$walltime[0],$mydate[0],$M{$mydate[1]},$mydate[2]-1900,);
    } elsif ($fw eq "f48" ) {
      my @walltime= ( $3, $4, $5 );
      my @mmydate= (localtime());
       my @mydate= @mmydate[3,4,5];
          @mydate[1]=$mydate[0];
          @mydate[0]=$M{$1};
          # assuming this year, could use a better algo...
          if ($1 > @mmydate[5] ) { @mydate[2]-- };
         $mynow=mktime($walltime[2],$walltime[1],$walltime[0],$mydate[0],$M{$mydate[1]},$mydate[2],);
    } elsif ($fw eq "acc" ) {
#$regexstart{'acc'}='^\S+\s+\S+\s+\S+\s+\[(\d+)\/(\S+)\/(\d+):(\d+):(\d+):(\d+)\s';
#204.101.134.87 - - [09/Nov/2004:14:06:57 -0500] "GET / HTTP/1.0" 200 641
      my @walltime= ( $4, $5, $6 );
      my @mydate= ( $1, $2, $3 );
         $mynow=mktime($walltime[2],$walltime[1],$walltime[0],$mydate[0],$M{$mydate[1]},$mydate[2]-1900,);
    }

    $lastdate=$mynow;
    if (!$firstdate) { 
          $firstdate=$mynow;
          $ifirstdate=$mynow;
          if ($align) {
            if ( $duration eq 'month') {
             # set firstdate to the beginning of the month with this time 
             $firstdate= timelocal(0,0,0,1,(localtime($firstdate))[4,5]);
            } else {
             # set firstdate to the beginning of the day with this time 
             $firstdate= timelocal(0,0,0,(localtime($firstdate))[3,4,5]);
            }
          }
    }
 }

########################################################################
# Main Loop
########################################################################

#my $fw="o35";
## For OpenBSD3.5
# OpenBSD3.5 pcap includes packet lengths, but the tcpdump utility does not
# display it.  If you update the tcpdump utility, then set fw to o36,
# and then byte counts should work.
$regex{'o35'}=
  '^(\d+)\.(\d+)\s+(\d+\.\d+\.\d+\.\d+)\.?(\d+)?\s+>\s+(\d+\.\d+\.\d+\.\d+)\.?(\d+)?:\s*(\w+|\d+)?(.*)';

$regex{'o35s'}=
  '^(\d+)\.(\d+)\s+rule (\S+):\s(\S+)\s(\S+)\son\s(\S+):\s(\d+\.\d+\.\d+\.\d+)\.?(\d+)?\s+>\s+(\d+\.\d+\.\d+\.\d+)\.?(\d+)?:\s*(\w+|\d+)?';

## For OpenBSD3.6
$regex{'o36'}=
  '^(\d+)\.(\d+)\s+(\d+\.\d+\.\d+\.\d+)\.?(\d+)?\s+>\s+(\d+\.\d+\.\d+\.\d+)\.?(\d+)?:\s*(\w+|\d+)?(.*)len\s+(\d+)';

$regex{'o36s'}=
  '^(\d+)\.(\d+)\s+rule (\S+):\s(\S+)\s(\S+)\son\s(\S+):\s(\d+\.\d+\.\d+\.\d+)\.?(\d+)?\s+>\s+(\d+\.\d+\.\d+\.\d+)\.?(\d+)?:\s*(\w+|\d+)?(.*)len\s+(\d+)';

#"7973724" "20Jul2005" "8:18:37" "XXXXXXXXXXXXXXXXXX" "ce0" "BlackHole" "Log" "Drop" "20589" "1.1.1.1"
# "2.2.2.2" "tcp" "" "80" "" "th_flags: 19; message_info: TCP packet out of state"
$regex{'csv1'}='^"([^"]*?)" "([^"]*?)" "([^"]*?)" "([^"]*?)" "([^"]*?)" "([^"]*?)" "([^"]*?)" "([^"]*?)" "([^"]*?)" "([^"]*?)" "([^"]*?)" "([^"]*?)" "([^"]*?)" "([^"]*?)" "([^"]*?)" "([^"]*?)"';

$regex{'f48'}=
'^(\S+)\s+(\d+)\s+\d+:\d+:\d+.*?:\s(\d+):(\d+):(\d+)\.(\d+)\s?(\d+)?x?\s+(\S+)\s(\S+)\s+(\S+)\s+([0-9.]+),(\d+)\s+->\s+([0-9.]+),(\d+)';
### Still need to address the len, etc, stuff
#'^(\S+)\s+(\d+)\s+\d+:\d+:\d+.*?:\s(\d+):(\d+):(\d+)\.(\d+)\s+(\S+)\s(\S+)\s+(\S+)\s+([0-9.]+),(\d+)\s+->\s+([0-9.]+),(\d+)\s+PR\s+(\S+)\s+len\s+(\d+)\s+(\d+)\s+(\S+)\s+(\S+)';
#                                                   msec     int    rule  st
#Aug 14 00:44:03 hurck ipmon[72]: 00:44:02.533431 2x xl0 @0:16 p 1.1.1.1,1027 -> 2.2.2.2,445 PR tcp len 20 48 -S IN 

#$regex{'acc'}='^(\S+)\s+(\S+)\s+(\S+)\s+\[(\d+)\/(\S+)\/(\d+):(\d+):(\d+):(\d+)\s+(\S+)\s+(\S+)\s+';
$regex{'acc'}='^(\S+)\s+(\S+)\s+(\S+)\s+\[(\d+)\/(\S+)\/(\d+):(\d+):(\d+):(\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s+';
#204.101.134.87 - - [09/Nov/2004:14:06:57 -0500] "GET / HTTP/1.0" 200 641

 
if ($state != 2 || $intf  || $nintf) { 
  if ($fw eq "o35" ) { $fw="o35s" }
  if ($fw eq "o36" ) { $fw="o36s" }
}

# This regex 
if (/$regex{$fw}/) {

 my $rt;  # The time of the record, in sec  
 my $rsa; # The source address for this record
 my $rsp; # The source port    for this record
 my $rda; # The dest   address for this record
 my $rdp; # The dest   port    for this record
 my $rp;  # The protocol of the record
 my $rhl; # The header length in bytes, or icmp code..., for this record
 my $len; # The packet length in bytes

 my $rrule; # rule number
 my $rstate;# pass or deny, etc.
 my $rdirec;# in our out
 my $rint;  # interface
 my $rmult;  # interface

 if ($fw eq "o36") {
  # Mapping of regex to used parameters
  #2=epoch micro
  $rt =$1;
  $rsa=$3;
  $rsp=$4;
  $rda=$5;
  $rdp=$6;
  $rp=$7; 
  $rhl=$8;
  $len=$9;
 } elsif ($fw eq "o36s") {
  # Mapping of regex to used parameters
  #2=epoch micro
  $rt =$1;
  $rrule=$3;
  $rstate=$4;
  $rdirec=$5;
  $rint=$6;
  $rsa=$7;
  $rsp=$8;
  $rda=$9;
  $rdp=$10;
  $rp=$11; 
  $rhl=$12;
  $len=$13;
 } elsif ($fw eq "o35") {
  # Mapping of regex to used parameters
  #2=epoch micro
  $rt =$1;
  $rsa=$3;
  $rsp=$4;
  $rda=$5;
  $rdp=$6;
  $rp=$7; 
  $rhl=$8;
 } elsif ($fw eq "o35s") {
  # Mapping of regex to used parameters
  #2=epoch micro
  $rt =$1;
  $rrule=$3;
  $rstate=$4;
  $rdirec=$5;
  $rint=$6;
  $rsa=$7;
  $rsp=$8;
  $rda=$9;
  $rdp=$10;
  $rp=$11; 
  } elsif ($fw eq "csv1") {
  $rrule=$13;
  $rstate=$8;
  $rint=$5;
  $rsa=$10;
  $rsp=$14;
  $rda=$11;
  $rdp=$9;
  if ( $fw == "csv1" ) {
      my $mydate=$2;
      my $walltime=$3;
      my @mydate = $mydate =~ /^(\d+)(\D+)(\d+)$/;
      my @walltime= $walltime =~ /^(\d+):(\d+):(\d+)$/;
      $rt=mktime($walltime[2],$walltime[1],$walltime[0],$mydate[0],$M{$mydate[1]},$mydate[2]-1900,);
  } 
 } elsif ($fw eq "f48") {
#'^(\S+)\s+(\d+)\s+\d+:\d+:\d+.*?:\s(\d+):(\d+):(\d+)\.(\d+)\s?(\d+)?x?\s+(\S+)\s(\S+)\s+(\S+)\s+([0-9.]+),(\d+)\s+->\s+([0-9.]+),(\d+)';
   $rmult=$6;
#  if ( $rmult !> 1 ) 
   $rint=$8;
   $rrule=$9;
   $rstate=$10;
#  $rdirec=;
   $rsa=$11;
   $rsp=$12;
   $rda=$13;
   $rdp=$14;
      my @walltime= ( $3, $4, $5 );
      my @mmydate= (localtime());
       my @mydate= @mmydate[3,4,5];
          @mydate[1]=$mydate[0];
          @mydate[0]=$M{$1};
          # assuming this year, could use a better algo...
          if ($1 > @mmydate[5] ) { @mydate[2]-- };
         $rt=mktime($walltime[2],$walltime[1],$walltime[0],$mydate[0],$M{$mydate[1]},$mydate[2],);
 } elsif ($fw eq "acc") {
#$regex{'acc'}='^(\S+)\s+(\S+)\s+(\S+)\s+\[(\d+)\/(\S+)\/(\d+):(\d+):(\d+):(\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s+';
#204.101.134.87 - - [09/Nov/2004:14:06:57 -0500] "GET / HTTP/1.0" 200 641
    $rsa=$1;
      my @walltime= ( $7, $8, $9 );
      my @mydate=   ( $4, $5, $6 );
         $rt=mktime($walltime[2],$walltime[1],$walltime[0],$mydate[0],$M{$mydate[1]},$mydate[2]-1900,);
         $rdp=$12;
         $rdp=~ s/^.*\///;
 } else {
  print STDERR "what fw type: $fw?\n";
  exit;
 }


#next if ( $state == 0 && $rstate ne "block");
#next if ( $state == 1 && $rstate ne "pass");
 next if ( $state == 0 && ( $rstate eq "pass" || $rstate eq "Accept" || $rstate eq "p") );
 next if ( $state == 1 && ( $rstate eq "block"||$rstate eq "Deny"||$rstate eq "Drop"||$rstate eq "d"));
 next if ( $intf  && $rint ne $intf);
 next if ( $nintf && $rint eq $nintf);

 my $x=1;          # $x=1, unless logs have multiplier, like syslog
#  $x=$rmult if ($rmult);
   $lastaddr=$rsa;

   $total{'count'}++;
   $total{'bytes'}+=$len;

   ${tgm{$rsa}{ &tgpopv($rt,$firstdate,$tgunit,"align") }}++; ###TG: populate TG hash

 my  $rport=$rdp?$rdp:$rp;  # $rport = destport or protocol
   if ($rport eq "icmp" )  {
     if    ($rhl =~ m/unreachable/) { $rport="icmpu" }
     elsif ($rhl =~ m/echo/       ) { $rport="icmpe" }
     elsif ($rhl =~ m/transit/    ) { $rport="icmpx" }
     elsif ($rhl =~ m/quench/     ) { $rport="icmpq" }
     elsif ($rhl =~ m/redirect/   ) { $rport="icmpr" }
     else { print "$rhl\n";}
   }

   $bysa{$rsa}{'count'}  +=$x;
   $bysa{$rsa}{'last'}    =$rt;
#  $bysa{$rsa}{'first'}?$bysa{$rsa}{'first'}:$rt;
   if ( ! $bysa{$rsa}{'first'} ) { $bysa{$rsa}{'first'}=$rt } ;
   $bysa{$rsa}{'Ips'}{$rport}+=$x;
   $bysa{$rsa}{'bytes'}  +=$len;
   $bysa{$rsa}{'Nets'}{$rda}++;

   if ($flagquiet == 0 || $flagquiet < -1) { ## populate by port info, by network info 
     my $srcnt=$rsa;
     $srcnt=~ s/\.[0-9]+$//;
     $bynet{$srcnt}{'count'}+=$x;
     $bynet{$srcnt}{'bytes'} +=$len;
     $bynet{$srcnt}{'last'}   =$rt;

     $byport{$rport}{'count'}+=$x;
     $byport{$rport}{'bytes'}+=$len;
     $byport{$rport}{'Nets'}{$srcnt}+=$x;
     $byport{$rport}{'Ips'}{$rsa}+=$x;
     $byport{$rport}{'last'}=$rt;
   }

   if ($flagquiet < 0 ) { ## populate srcaddr.dstaddr.dstport info
     my $myquad=sprintf("%-15s -%-15s:%-6s",$rsa,$rda,$rport);
     $byquad{$myquad}{'count'}++;
     $byquad{$myquad}{'last'}=$rt;
     ###TG: populate TG hash
     ${tgq{$myquad}{ &tgpopv($rt,$firstdate,$tgunit,"align") }}++; 
   }

   $bysa{$rsa}{'rfc1918'}=1 if ( $rda !~ /^(10\.|172\.(1[6-9]|2[0-9]|3[0-2])\.|192\.168\.)/ ) ;

 } elsif (/echoreply.*OUT\s*$/) {
   #noop;
 } elsif ( $flagsigs && /^\t\t\t\s../ ) {
   foreach my $sig ( keys %sigs ) {
     if ( /$sigs{$sig}/ ) { $bysa{$lastaddr}{'Ips'}{$sig}=1 ;
#     printf "S:%s: %s: %s\n", $sig, $sigs{$sig}, $_;
     }
   }
   #noop;
 } else {
   print STDERR "ERR: $_";
 }#if-ipmon
}#while reading input

########################################################################
# Calculate some things and start printing
########################################################################
 $ifirstdate = int($ifirstdate);
 $lastdate = int($lastdate);
 $ENV{'TZ'}='UTC';
 my $f1=&epoch2isotime ($ifirstdate);
 my $d1=&epoch2isotime ($lastdate);
 $ENV{'TZ'}='Canada/Eastern';
 my $f2=&epoch2isotime ($ifirstdate);
 my $d2=&epoch2isotime ($lastdate);

printf ("Args: %s\n",$args) if ($args);

my $stringlast="";
   $stringlast=" (last ".$inlastsecs." seconds)" if ($inlastsecs);
   $stringlast=" (scale: $scale)" if ($scale != 1);
printf ("Date: %s - %s%s\n",$ifirstdate,$lastdate,$stringlast);
printf ("Date: %s - %s GMT\n",$f1,$d1);
printf ("Date: %s - %s Eastern\n",$f2,$d2);

if ($archive) { $now = $lastdate }


if ($flagquiet == 0 || $flagquiet < -1) {
 my $n=0;
 printf ("\nPort Summary - top accessed ports:\n");
 printf ("%6s:%6s %5s %5s %8s\n", "count","port","u.addr","u.net","bytes");
 foreach my $i ( sort { $byport{$b}{'count'} <=> $byport{$a}{'count'} } keys %byport) {
    next if ( $inlastsecs && $now - $byport{$i}{'last'} > $inlastsecs );
    my $ips=0; my $nets=0;
    foreach my $k ( sort keys %{ $byport{$i}{'Ips'}  } ) { $ips++  }
    foreach my $k ( sort keys %{ $byport{$i}{'Nets'} } ) { $nets++ }
    printf ("%6s: %6s %5s %5s %8s\n", $byport{$i}{'count'},$i,$ips,$nets,$byport{$i}{'bytes'});
    if ($n++ >= $nthresh - 1 ) { 
     print "       overflow\n";
     last ;
     };
 }
}

my $E;
my $G;

if ( $duration eq 'month' ) {
   if ($align) {
    $G="         11111111112222222222333";
    $E="12345678901234567890123456789012";
   } else {
    $G="                                ";
    $E="                                ";
   }
} else {
   if ($align) {
    $G="5678901n12345678901M1234G";
    $E="012345678901n12345678901E";
    if ((localtime($firstdate))[8]) {
      $G="45678901n12345678901M123G";
    }
   } else {
      $G="         1111111111222222";
      $E="1234567890123456789012345";
   }
}

$n=0;
printf ("\nSource and Port Plus  (%s bytes  %s packets)\n",$total{'bytes'}, $total{'count'});
printf ("%6s:%16s %25s %s\n", "count","source",$G, " bytes  dstn  some ports");
printf ("%6s:%16s %25s %s %s\n", "-----","------",$E,"------"," ----  -----");
foreach my $i ( sort { $bysa{$b}{'count'} <=> $bysa{$a}{'count'} } keys %bysa) {
   next if ( $inlastsecs && $now - $bysa{$i}{'last'} > $inlastsecs );
   next if ( $diff && $bysa{$i}{'last'} - $bysa{$i}{'first'} < $diff );
   my @list=undef;
   my @dstn=values %{ $bysa{$i}{'Nets'} };
#  push (@list,$#list.":") ;
   my $tg;
   foreach my $j (0..$tgunits){$tg=$tg.&tgrn($tgm{$i}{$j})}###TG: build
   my $ips=0;
   foreach my $k (sort keys %{ $bysa{$i}{'Ips'} }) {$ips++;push(@list,$k)}
   printf ("%6s: %-15s %25s", $bysa{$i}{'count'},$i,$tg);
   my $m=0;
   printf "%7s", $bysa{$i}{'bytes'};
   printf " %3s", $#dstn+1;
   printf " %1s", $bysa{$i}{'rfc1918'}?"N":" "  ;
   printf " %s:", $#list;
   foreach my $k ( @list ) { print " $k"; if ($m++>=$nthresh) { last} }
   print "\n";
   if ($n++ >= $ntop - 1 ) { 
    my $rc=0;
    for my $c ( keys %bysa) { $rc++ }
    printf "       overflow (%s of %s)\n", $ntop, $rc;
    #rint "       overflow\n";
    last ;
   };
}
 
if ( $flagquiet == 0 || $flagquiet < -1) {
 my $n=0;
 printf ("\nNetwork Summary:\n");
 foreach my $i (sort { $bynet{$b}{'count'} <=> $bynet{$a}{'count'} } keys %bynet) {
    next if ( $inlastsecs && $now - $bynet{$i}{'last'} > $inlastsecs );
    printf ("%6s: %-12s %10s\n", $bynet{$i}{'count'},$i, $bynet{$i}{'bytes'});
    if ($n++ >= $ntop - 1 ) { 
     print "       overflow\n";
     last ;
    };
 }
}

if ( $flagquiet < 0 ) {
 my $n=0;
 printf ("Connection Summary: %40s\n","start...end");
 foreach my $i (sort { $byquad{$b}{'count'} <=> $byquad{$a}{'count'} } keys %byquad) {
    next if ( $inlastsecs && $now - $byquad{$i}{'last'} > $inlastsecs );
    my $tg;
    foreach my $j (0..$tgunits){$tg=$tg.&tgrn($tgq{$i}{$j})}###TG: build
    printf ("%6s:%12s %12s\n", $byquad{$i}{'count'},$i,$tg);
    if ($n++ >= $ntop - 1 ) { 
     print "       overflow\n";
     last 
    };
 } 
} # 

########################################################################
sub epoch2isotime () {
 my @rawdate;
 @rawdate=localtime($_[0]);
 return sprintf("%02s/%02s/%02s %02s:%02s:%02s%s",
                 $rawdate[5]+1900,$rawdate[4]+1,$rawdate[3],
                 $rawdate[2],$rawdate[1],$rawdate[0],
                 $rawdate[8]?" D":" S") ;
}

########################################################################

########################################################################
## TG: text-graph routines
########################################################################
# The intent of this part is to setup 'standard' text graphs.
#
# The following would be an example text-graph:
#   "4CLXX9841..2"
# Each char, or column, represents a time period.
#  The time period could be 1 day, in which case this graph would
#  show a 12 day period.  If the time period was 30 days, then the
#  graph would show 360 days, or about 1 year.
# Each char represents the count of 'hits' found withing the
#  time period.  So, for instance, the first time period here
#  had 4 hits.  The second time period has a 'C'.  Because we
#  only have a single char, I've chose to use Roman Numerals to
#  represent numbers over 9. 
#    X=10
#    L=50
#    C=100
#  The period represents 0.  It is much easier to view the text
#  graph with the 0's repaced with periods.
#
# Currently there is no 'alignment' option.  This means that the
# the first time period begins the last second for which we have
# data, and the second last period begins exactly 1 time period
# before the last second.  In the future there may be an alignment
# option.  The alignment option would align the time periods with
# one's understanding of how they should be aligned to a wall clock.
# For instance, if the time period is 24 hours, it would be best
# if the each period was aligned with a day as seen on the clock
# on the wall, as opposed to a day being 60*60*24 seconds previous
# to the last time sample available.  More on this later.
#
#
# The 'standard' text garph will include
#  1. a 6 char field, each char representing 10 seconds
#  2. a 6 char field, each char representing 10 minutes
#  3. a 12 char field, each char representing 2 hours
#  4. a 7 char field, each char representing  1 day
#  4. a 8 char field, each char representing  1 week
#  5. a 12 char field, each char representing 30 days
# but you are free to set other non-standard graphs
# Perhaps all char fields should be set to 12?
########################################################################
 # This routine takes in a number and returns a single char
 # which represents a number.  uses roman numerals for values > 9
 #######################################################################
sub tgrn () {
   if (! defined($_[0])) {return "."}
   return  
           $_[0]/$scale>=1000 ? "M": 
           $_[0]/$scale>=100  ? "C": 
           $_[0]/$scale>=50   ? "L": 
           $_[0]/$scale>=10   ? "X": 
           $_[0]/$scale>=1    ? int($_[0]/$scale): 
           $_[0]>=1           ? "+": 
                                ".";
}
########################################################################
 # This prints the Time period, etc, used.
 #######################################################################
sub tgprintsettings() {
  printf "Duration %s\n",$_[0];
}
########################################################################
 # This sets the time period length and number, and provides a string
 #######################################################################
 # perhaps all should be set to 12 time periods?
sub tgautoset() {
  if ( $_[0] <=60 ) {return(10,6,"graph units=10 seconds")
  }elsif($_[0]<=60*60) {return(60*10,6,"graph units=10 minutes")
  }elsif($_[0]<=60*60*24) {return(60*60*2,12,"graph units=2 hours")
  }elsif($_[0]<=60*60*24*7) {return(60*60*24,7,"graph units=1 day")
  }elsif($_[0]<=60*60*24*10) {return(60*60*24,10,"graph units=1 day")
  }elsif($_[0]<=60*60*24*7*8) {return(60*60*24*7,8,"graph units=1 week")
  }elsif($_[0]<=60*60*24*30*12){return(60*60*24*30,12,"graph units=30 days")
  }else { return (9999999,1,"graph units failed")
  }
}
########################################################################
 # This calculates which time period in the hash to populate
 # This routine should be updated in order to do 'alignment'
 # it would require a reference, or offset, or some other argument
 # which currenty does not exit, maybe.  (it could be calculated
 # here, but this routine may get called too often for that to make
 # sense.)
 #######################################################################
sub tgpopv(){
 # one means to do the alignment would be to *add*
 # the offset of the alignment (we should add another period to
 # our output of we do this, though, sort of.
 # warning: off-by-one errors, not tested, etc.
 if (! defined($_[3])) {
  return int(($_[0]-$_[1])/$_[2] ) ; 
 } else {
  # this has not yet been tested
# printf "XX:%s\n", $_[0]%$_[2];
##return int(($_[0]-$_[1]+ $_[0]%$_[2])/$_[2] ) ; #only gmt/utc?
  return int(($_[0]-$_[1])/$_[2] ) ; 
 } 
}
###TG###################################################################
